#!/usr/bin/env python
# -*- coding: utf-8 -*-

# py_tools_ds - A collection of geospatial data analysis tools that simplify standard
# operations when handling geospatial raster and vector data as well as projections.
#
# Copyright (C) 2016-2021
# - Daniel Scheffler (GFZ Potsdam, daniel.scheffler@gfz-potsdam.de)
# - Helmholtz Centre Potsdam - GFZ German Research Centre for Geosciences Potsdam,
#   Germany (https://www.gfz-potsdam.de/)
#
# This software was developed within the context of the GeoMultiSens project funded
# by the German Federal Ministry of Education and Research
# (project grant code: 01 IS 14 010 A-C).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
test_coord_trafo
----------------

Tests for `py_tools_ds.geo.coord_trafo` module.
"""

from unittest import TestCase, main
from shapely.geometry import Polygon
import numpy as np

from py_tools_ds.geo.projection import EPSG2WKT
from py_tools_ds.geo.coord_trafo import (
    imXY2mapXY, mapXY2imXY,
    pixelToLatLon, latLonToPixel,
    reproject_shapelyGeometry,
    transform_any_prj,
    get_utm_zone
)

poly_local = Polygon([(5708, -3006),
                      (5708, -3262),
                      (5452, -3262),
                      (5452, -3006),
                      (5708, -3006)])

ll_x, ll_y = (10.701359,
              47.54536)
utm_x, utm_y = (628023.1302739962,
                5267173.503313034)

geotransform_utm = (331185.0, 30.0, -0.0, 5840115.0, -0.0, -30.0)
wkt_utm = \
    """
    PROJCS["WGS 84 / UTM zone 33N",
           GEOGCS["WGS 84",
                  DATUM["WGS_1984",
                        SPHEROID["WGS 84", 6378137, 298.257223563,
                                 AUTHORITY["EPSG", "7030"]],
                        AUTHORITY["EPSG", "6326"]],
                  PRIMEM["Greenwich", 0,
                         AUTHORITY["EPSG", "8901"]],
                  UNIT["degree", 0.0174532925199433,
                       AUTHORITY["EPSG", "9122"]],
                  AUTHORITY["EPSG", "4326"]],
           PROJECTION["Transverse_Mercator"],
           PARAMETER["latitude_of_origin", 0],
           PARAMETER["central_meridian", 15],
           PARAMETER["scale_factor", 0.9996],
           PARAMETER["false_easting", 500000],
           PARAMETER["false_northing", 0],
           UNIT["metre", 1,
                AUTHORITY["EPSG", "9001"]],
           AXIS["Easting", EAST],
           AXIS["Northing", NORTH],
           AUTHORITY["EPSG", "32633"]]
    """
wkt_utm = ' '.join(wkt_utm.split())


class Test_reproject_shapelyGeometry(TestCase):

    def test_reproject_shapelyGeometry(self):
        poly_lonlat = reproject_shapelyGeometry(poly_local, 32636, 4326)
        self.assertTrue(isinstance(poly_lonlat, Polygon))


class Test_transform_any_prj(TestCase):

    def test_lonlat_to_utm(self):
        out_x, out_y = transform_any_prj(4326, 32632, ll_x, ll_y)
        self.assertTrue(isinstance(out_x, float))
        self.assertTrue(isinstance(out_y, float))
        self.assertTrue(np.all(np.isclose(np.array([out_x, out_y]),
                                          np.array([utm_x, utm_y]))))


def get_utm_gt_prj_from_lonlat(lon, lat):
    utmepsg = int(f"32{6 if lat > 0 else 7}{get_utm_zone(lon)}")
    utmx, utmy = transform_any_prj(4326, utmepsg, lon, lat)
    gt = (utmx, 30, 0, utmy, 0, -30)
    prj = EPSG2WKT(utmepsg)

    return gt, prj


class Test_mapXY2imXY(TestCase):

    def test_singlecoord_tuple_at_origin(self):
        out = mapXY2imXY((331185.0, 5840115.0), geotransform_utm)
        self.assertIsInstance(out, tuple)
        self.assertEqual(out, (0, 0))

    def test_singlecoord_tuple_at_1_1(self):
        out = mapXY2imXY((331215.0, 5840085.0), geotransform_utm)
        self.assertIsInstance(out, tuple)
        self.assertEqual(out, (1, 1))

    def test_singlecoord_nparray_at_1_1(self):
        out = mapXY2imXY(np.array([331215.0, 5840085.0]), geotransform_utm)
        self.assertIsInstance(out, np.ndarray)
        self.assertTrue(np.array_equal(out, np.array([1, 1])))

    def test_multicoords_nparray(self):
        out = mapXY2imXY(np.array([[331185.0, 5840115.0],
                                   [331215.0, 5840085.0]]), geotransform_utm)
        self.assertIsInstance(out, np.ndarray)
        self.assertTrue(np.array_equal(out, np.array([[0, 0], [1, 1]])))

    def test_tuple_ndarray_outputConsistency(self):
        tupleout = mapXY2imXY((331215.0, 5840085.0), geotransform_utm)
        npout = mapXY2imXY(np.array([331215.0, 5840085.0]), geotransform_utm)
        self.assertTrue(np.array_equal(np.array(tupleout),
                                       npout))


class Test_imXY2mapXY(TestCase):

    def test_singlecoord_tuple_at_origin(self):
        out = imXY2mapXY((0, 0), geotransform_utm)
        self.assertIsInstance(out, tuple)
        self.assertEqual(out, (331185.0, 5840115.0))

    def test_singlecoord_tuple_at_1_1(self):
        out = imXY2mapXY((1, 1), geotransform_utm)
        self.assertIsInstance(out, tuple)
        self.assertEqual(out, (331215.0, 5840085.0))

    def test_singlecoord_nparray_at_1_1(self):
        out = imXY2mapXY(np.array([1, 1]), geotransform_utm)
        self.assertIsInstance(out, np.ndarray)
        self.assertTrue(np.array_equal(out, np.array([331215.0, 5840085.0])))

    def test_multicoords_nparray(self):
        out = imXY2mapXY(np.array([[0, 0], [1, 1]]), geotransform_utm)
        self.assertIsInstance(out, np.ndarray)
        self.assertTrue(np.array_equal(out, np.array([[331185.0, 5840115.0],
                                                      [331215.0, 5840085.0]])))

    def test_tuple_ndarray_outputConsistency(self):
        tupleout = imXY2mapXY((1, 1), geotransform_utm)
        npout = imXY2mapXY(np.array([1, 1]), geotransform_utm)
        self.assertTrue(np.array_equal(np.array(tupleout),
                                       npout))


class Test_pixelToLatLon(TestCase):

    def test_standard_case(self):
        latlonPairs = pixelToLatLon([(0, 0)], geotransform=geotransform_utm, projection=wkt_utm)
        self.assertIsInstance(latlonPairs, list)
        self.assertEqual(len(latlonPairs), 1)

    def test_negative_imcoords(self):
        latlonPairs = pixelToLatLon([(-10, -10)], geotransform=geotransform_utm, projection=wkt_utm)
        self.assertIsInstance(latlonPairs, list)
        self.assertEqual(len(latlonPairs), 1)

    def test_negative_lon(self):
        for lon, lat in [(-1, 30), (-89, -30), (-90, -30), (-91, 30), (-130, -30), (-79.999999, -30)]:
            latlonPairs = pixelToLatLon([(0, 0)], *get_utm_gt_prj_from_lonlat(lon=lon, lat=lat))
            self.assertTrue(np.allclose(np.array(latlonPairs),
                                        np.array([(lat, lon)])),
                            msg=f"{latlonPairs} != {[(lat, lon)]}")

    def test_negative_lat(self):
        for lon, lat in [(-1, -1), (1, -30), (-1, -45), (1, -89.999999)]:
            latlonPairs = pixelToLatLon([(0, 0)], *get_utm_gt_prj_from_lonlat(lon=lon, lat=lat))
            self.assertTrue(np.allclose(np.array(latlonPairs),
                                        np.array([(lat, lon)])),
                            msg=f"{latlonPairs} != {[(lat, lon)]}")

    def test_position_out_of_bounds(self):
        with self.assertRaises(RuntimeError):
            pixelToLatLon([(0, 0)], (0, 1, 0, -91, 0, -1), EPSG2WKT(4326))
        with self.assertRaises(RuntimeError):
            pixelToLatLon([(0, 0)], (361, 1, 0, 0, 0, -1), EPSG2WKT(4326))

    def test_multiple_coords(self):
        latlonPairs = pixelToLatLon([(-10, -10), (10, 10)], geotransform=geotransform_utm, projection=wkt_utm)
        self.assertIsInstance(latlonPairs, list)
        self.assertEqual(len(latlonPairs), 2)


class Test_latLonToPixel(TestCase):

    def test_default_case(self):
        lat, lon = 51.1234, 12.1234
        pixXYPairs = latLonToPixel([(lat, lon)], *get_utm_gt_prj_from_lonlat(lon, lat))
        self.assertIsInstance(pixXYPairs, list)
        self.assertEqual(len(pixXYPairs), 1)
        self.assertEqual(pixXYPairs, [[0, 0]])

    def test_negative_lon(self):
        lat, lon = 51.1234, -12.1234
        pixXYPairs = latLonToPixel([(lat, lon)], *get_utm_gt_prj_from_lonlat(lon, lat))
        self.assertIsInstance(pixXYPairs, list)
        self.assertEqual(len(pixXYPairs), 1)
        self.assertEqual(pixXYPairs, [[0, 0]])

    def test_negative_lat(self):
        lat, lon = -51.1234, 12.1234
        pixXYPairs = latLonToPixel([(lat, lon)], *get_utm_gt_prj_from_lonlat(lon, lat))
        self.assertIsInstance(pixXYPairs, list)
        self.assertEqual(len(pixXYPairs), 1)
        self.assertEqual(pixXYPairs, [[0, 0]])

    def test_lon_out_of_bounds(self):
        lat, lon = 51.1234, 361
        with self.assertRaises(ValueError):
            latLonToPixel([(lat, lon)],
                          (181, 1, 0, 51.1234, 0, -1),
                          EPSG2WKT(int(f"32{6 if lat > 0 else 7}{get_utm_zone(181)}")))

    def test_lat_out_of_bounds(self):
        lat, lon = -91, 12.1234
        with self.assertRaises(ValueError):
            latLonToPixel([(lat, lon)],
                          (12.1234, 1, 0, -91, 0, -1),
                          EPSG2WKT(int(f"32{6 if lat > 0 else 7}{get_utm_zone(lon)}")))


if __name__ == '__main__':
    main()
